/*! \file rhs.h
 *  \brief Implements right hand side assembly for exercise 1
 */

#ifndef __FORCESSHELLFE_H
#define __FORCESSHELLFE_H

#include <cmath>

#include <discreteFunctionShellFE.h>
#include <shellHandler.h>
#include <unityTriangleIntegratorShellFE.h>

namespace shellFE {

//! Assembly operator for constant RHS
template< typename ConfiguratorType >
class createConstantRHS :
    public UnitTriangleFENonlinOpIntegratorShellFE <ConfiguratorType, createConstantRHS <ConfiguratorType> >
{
protected:
    typedef typename ConfiguratorType::RealType RealType;
    typedef typename ConfiguratorType::TangentVecType TangentVecType;

    const RealType _val;

public:
    createConstantRHS(const ConfiguratorType &conf, const RealType val)
        : UnitTriangleFENonlinOpIntegratorShellFE <ConfiguratorType, createConstantRHS <ConfiguratorType> > (conf),
          _val(val)
    {}

    RealType getNonlinearity ( const typename ConfiguratorType::ElementType &el, int /*quadPoint*/) const
    {
        return 2.0 * el.getAreaOfFlattenedTriangle() * _val;
    }
};

//! Assembly operator for non-constant RHS
template< typename ConfiguratorType >
class createNonConstantRHS :
    public UnitTriangleFENonlinOpIntegratorShellFE <ConfiguratorType, createNonConstantRHS <ConfiguratorType> >
{
protected:
    typedef typename ConfiguratorType::RealType RealType;
    typedef typename ConfiguratorType::TangentVecType TangentVecType;

    static const int dim = 3;

    std::unique_ptr<typename ConfiguratorType::VectorType> _dof;
    std::unique_ptr<DiscreteVectorFunctionDefaultShellFE<ConfiguratorType>> _x;

public:
    createNonConstantRHS(const ConfiguratorType &conf)
        : UnitTriangleFENonlinOpIntegratorShellFE <ConfiguratorType, createNonConstantRHS <ConfiguratorType> > (conf)
    {
        int numGlobalDofs = this->_config.getNumGlobalDofs();
        const typename ConfiguratorType::InitType &grid = this->_config.getInitializer();
        _dof.reset(new typename ConfiguratorType::VectorType (dim * numGlobalDofs));

        for (int i = 0; i < numGlobalDofs; ++i) {
            for (int c = 0; c < dim; ++c) {
                (*_dof)[c*numGlobalDofs + i] = grid.getVertex(i)[c];
            }
        }

        _x.reset(new DiscreteVectorFunctionDefaultShellFE<ConfiguratorType>(this->_config, *_dof, dim));
    }

    RealType getNonlinearity ( const typename ConfiguratorType::ElementType &el, int q) const
    {
        const typename ConfiguratorType::BaseFuncSetType &bfs = this->_config.getBaseFunctionSet(el);
        typename ConfiguratorType::BaseFuncSetType::DomVecType coords = bfs.getRefCoord(q);

        typename ConfiguratorType::TangentVecType cartCoord;
        _x->evaluateAtQuadPoint(el, q, cartCoord);

        RealType nl = 2.0 * el.getAreaOfFlattenedTriangle() * evalRHS(cartCoord);

        return nl;
    }

    RealType evalRHS(typename ConfiguratorType::TangentVecType &cartCoord) const
    {
        RealType nl = std::pow(aol::Sqr(-1. + 2. * cartCoord[0]) + aol::Sqr(-1. + 2. * cartCoord[1]), 1./3.) *
                sin(2./3. * atan2(-1. + 2.*cartCoord[1], -1. + 2.*cartCoord[0]));

        return nl;
    }
};

//! \brief Assembles constant rhs
template <typename ConfiguratorType>
void createMaskedConstantRHS(const ConfiguratorType &conf,
                             const ShellHandler<ConfiguratorType> &shellHandler,
                             typename ConfiguratorType::RealType val,
                             typename ConfiguratorType::VectorType &rhs,
                             const bool collapseBoundaryValues = true)
{
    const typename ConfiguratorType::MaskType boundaryMask(
        shellHandler.getDirichletMask());
    int numGlobalDofs = conf.getNumGlobalDofs();
    rhs.setZero();

    createConstantRHS<ConfiguratorType>(conf, val).assembleAdd(rhs);

    //bc rhs_force
    if (collapseBoundaryValues) {
        for (int i = 0; i < numGlobalDofs; ++i) {
            if (boundaryMask[i]) {
                rhs[i] = 0.0;
            }
        }
    }
}

//! \brief Assembles sin-type rhs
template <typename ConfiguratorType>
void createMaskedNonConstantRHS(const ConfiguratorType &conf,
                                const ShellHandler<ConfiguratorType> &shellHandler,
                                typename ConfiguratorType::RealType val,
                                typename ConfiguratorType::VectorType &rhs,
                                const bool collapseBoundaryValues = true)
{
    const typename ConfiguratorType::MaskType boundaryMask(
        shellHandler.getDirichletMask());
    int numGlobalDofs = conf.getNumGlobalDofs();
    rhs.setZero();

    createNonConstantRHS<ConfiguratorType>(conf).assembleAdd(rhs);

    //bc rhs_force
    if (collapseBoundaryValues) {
        for (int i = 0; i < numGlobalDofs; ++i) {
            if (boundaryMask[i]) {
                rhs[i] = 0.0;
            }
        }
    }
}

} // end namespae

#endif
